/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2004/08/17 09:46:56 $ $Revision: 1.10 $ */
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <X11/xpm.h>
#include "global.h"
#include "Texture/Ball.xpm"
#include "Texture/Ball2.xpm"
#include "Texture/Wall.xpm"
#include "Texture/Wall2.xpm"
#include "Texture/Ball3.xpm"
#include "Texture/Ball4.xpm"
#include "Texture/Ball5.xpm"
#include "Texture/Ball6.xpm"
#include "Texture/Ball7.xpm"
#include "Texture/Ball8.xpm"

/* prototypes */
void perr(char *, ...);
void Redraw_Light(void);

/* quick and dirty bitmap loader...for 24 bit bitmaps with 1 plane only. */
int ImageLoadFromBmp(char *filename, MGLtexdetail *image) {
    FILE *file;
    /* size of the image in bytes. */
    unsigned long size;
    unsigned long i;
    /* number of planes in image(must be 1) */
    unsigned short int planes;
    /* number of bits per pixel (must be 24) */
    unsigned short int bpp;
    char temp;
    /* temporary color storage for bgr-rgb conversion. */

    /* make sure the file is there. */
    if ((file = fopen(filename, "rb"))==NULL)
    {
	printf("File Not Found : %s\n",filename);
	return 0;
    }
    
    /* seek through the bmp header, up to the width/height: */
    fseek(file, 18, SEEK_CUR);

    /* read the width */
    if ((i = fread(&image->width, 4, 1, file)) != 1) {
	printf("Error reading width from %s.\n", filename);
	return 0;
    }
    
    /* read the height */
    if ((i = fread(&image->height, 4, 1, file)) != 1) {
	printf("Error reading height from %s.\n", filename);
	return 0;
    }
    
    /* calculate the size (assuming 24 bits or 3 bytes per pixel). */
    size = image->width * image->height * 3;

    /* read the planes */
    if ((fread(&planes, 2, 1, file)) != 1) {
	printf("Error reading planes from %s.\n", filename);
	return 0;
    }
    if (planes != 1) {
	printf("Planes from %s is not 1: %u\n", filename, planes);
	return 0;
    }

    /* read the bpp */
    if ((i = fread(&bpp, 2, 1, file)) != 1) {
	printf("Error reading bpp from %s.\n", filename);
	return 0;
    }
    if (bpp != 24) {
	printf("Bpp from %s is not 24: %u\n", filename, bpp);
	return 0;
    }
	
    /* seek past the rest of the bitmap header.*/
    fseek(file, 24, SEEK_CUR);

    /* read the data. */
    image->data = (unsigned char *) malloc(size);
    if (image->data == NULL) {
	printf("Error allocating memory for color-corrected image data");
	return 0;	
    }

    if ((i = fread(image->data, size, 1, file)) != 1) {
	printf("Error reading image data from %s.\n", filename);
	return 0;
    }
    /* reverse all of the colors. (bgr -> rgb) */
    for (i=0;i<size;i+=3) {
	temp = (char) image->data[i];
	image->data[i] = image->data[i+2];
	image->data[i+2] = temp;
    }
    return 1;
}

unsigned char ShortConvertHex(char *s,int l)
{
  unsigned char val;
  #define  hex(k) (tolower(k)<'a' ? k-'0': tolower(k)-'a'+10)

  if (l==0) return(0);
  val=hex(s[0])*16;
  if (l>1) val+=hex(s[1]);
  return (val);
}

void ReadXpmToTexture(char **xpmdata,MGLtexdetail *texdet)
{
  int a,b,l;
  XpmImage image;
  XpmInfo info;
  unsigned char *colors[3];
  char *s;

  XpmCreateXpmImageFromData(xpmdata,&image,&info);
  texdet->width=image.width;
  texdet->height=image.height;
  texdet->components=3;
  texdet->level=0;
  texdet->border=0;
  texdet->format=GL_RGB;
  texdet->type=GL_UNSIGNED_BYTE;
  texdet->data=(unsigned char *)
    	calloc(texdet->width*texdet->height*texdet->components,sizeof(char));
  if (texdet->data==NULL) perr("malloc failed (data,ReadXpmToTexture)");
  for (a=0;a<3;a++) {
    colors[a]=(unsigned char *) calloc(image.ncolors,sizeof(char));
  }
  if (colors[2]==NULL) perr("malloc failed (colors,ReadXpmToTexture)");
  /* convert colors */
  for (a=0;a<(int)image.ncolors;a++) {
    s=image.colorTable[a].c_color;
    if (s!=NULL) {
      if (s[0]=='#') {
	l=strlen(s);
	for (b=0;b<3;b++) {
	  colors[b][a]=ShortConvertHex(&s[(l-1)/3*b+1],(l-1)/3);
	}
      } else { /* try to convert few named colors */
	if (strcasecmp(s,"None")==0) {
          colors[0][a]=colors[1][a]=colors[2][a]=0;
	} else if (strcasecmp(s,"black")==0) {
          colors[0][a]=colors[1][a]=colors[2][a]=0;
	} else if (strcasecmp(s,"white")==0) {
          colors[0][a]=colors[1][a]=colors[2][a]=255;
	} else if (strcasecmp(s,"red")==0) {
          colors[0][a]=255;colors[1][a]=colors[2][a]=0;
	} else if (strcasecmp(s,"green")==0) {
          colors[1][a]=255;colors[0][a]=colors[2][a]=0;
	} else if (strcasecmp(s,"blue")==0) {
          colors[2][a]=255;colors[1][a]=colors[0][a]=0;
	} else if (strcasecmp(s,"magenta")==0) {
          colors[1][a]=0;colors[2][a]=colors[0][a]=255;
	} else if (strcasecmp(s,"cyan")==0) {
          colors[0][a]=0;colors[2][a]=colors[1][a]=255;
	} else if (strcasecmp(s,"yellow")==0) {
          colors[1][a]=0;colors[0][a]=colors[1][a]=255;
	} else {
	  fprintf(stderr,"This stupid converter supports #rgb ");
	  fprintf(stderr,"and the 8 main colornames, using black...\n");
	  colors[0][a]=colors[1][a]=colors[2][a]=0;
	}
      }
    } else {
      fprintf(stderr,"This stupid converter supports only c color field in");
      fprintf(stderr," XPM files, using black...\n");
      colors[0][a]=colors[1][a]=colors[2][a]=0;
    }
  }
  for (a=0;a<(int)image.width*(int)image.height;a++) {
    for (b=0;b<3;b++) {
      texdet->data[a*3+b]=colors[b][image.data[a]];
    }
  }
  for (b=0;b<3;b++) free(colors[b]);
}


/* Load Bitmaps And Convert To Textures */
void LoadGLTextures() {	
    int a;

    ReadXpmToTexture(XPMWall,&gl.texture.detail[0]);
    ReadXpmToTexture(XPMBall,&gl.texture.detail[1]);
    ReadXpmToTexture(XPMBall2,&gl.texture.detail[2]);
    ReadXpmToTexture(XPMWall2,&gl.texture.detail[3]);
    ReadXpmToTexture(XPMBall3,&gl.texture.detail[4]);
    ReadXpmToTexture(XPMBall4,&gl.texture.detail[5]);
    ReadXpmToTexture(XPMBall5,&gl.texture.detail[6]);
    ReadXpmToTexture(XPMBall6,&gl.texture.detail[7]);
    ReadXpmToTexture(XPMBall7,&gl.texture.detail[8]);
    ReadXpmToTexture(XPMBall8,&gl.texture.detail[9]);
/*    if (!ImageLoadFromBmp("Texture/Ball.bmp", &gl.texture.detail[0])) {
	exit(1);
    }        
    if (!ImageLoadFromBmp("Texture/Wall.bmp", &gl.texture.detail[1])) {
	exit(1);
    }        */

    /* Create Texture */
    glGenTextures(gl.texture.num, gl.texture.texture);
    for (a=0;a<gl.texture.num;a++) {
      glBindTexture(GL_TEXTURE_2D, gl.texture.texture[a]);

    /* 2d texture, level of detail 0 (normal), 3 components (red,
     * green, blue), x size from image, y size from image, border 0
     * (normal), rgb color data, unsigned byte data, and finally the
     * data itself.*/
      glTexImage2D(GL_TEXTURE_2D, 0, 3, gl.texture.detail[a].width,
	gl.texture.detail[a].height, 0,
	GL_RGB, GL_UNSIGNED_BYTE, gl.texture.detail[a].data);

    /* Scale lineraly when texture size does not match */
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    }
}

/* initializes gl environment to some usefull state */
void reset_gl(GtkWidget *widget)
{
  int a,b,c;
  static GLfloat deflight[4][4] = {{ 0.0, 0.0, 5.0, 0.0},
    				    {0.3, 0.3, 0.3, 0.3},
				    {1.0, 1.0, 1.0, 1.0},
				    {0.0, 0.0, 0.0, 1.0}};

  /* light*/
  for (a=0;a<MaxLight;a++) {
    gl.light[a].on=0;
    gl.light[a].light=GL_LIGHT0+a;
    gl.light[a].type[0]=GL_POSITION;
    gl.light[a].type[1]=GL_AMBIENT;
    gl.light[a].type[2]=GL_DIFFUSE;
    gl.light[a].type[3]=GL_SPECULAR;
    for (c=0;c<MaxLightParam;c++) {
      for (b=0;b<4;b++) {
	gl.light[a].param[c][b]=deflight[c][b];
      }
    }
  }
  gl.light[0].on=1;
  gl.lightmodel.model=1;
  gl.lightmodel.modelname[0]=GL_LIGHT_MODEL_TWO_SIDE;
  gl.lightmodel.modelparam[0][0]=1.0;
  /* texture */
  gl.texture.num=10;
  gl.texture.texture=(GLuint *) calloc(gl.texture.num,
      			sizeof(GLuint));
  gl.texture.detail=(MGLtexdetail *) calloc(gl.texture.num,
      			sizeof(MGLtexdetail));
  LoadGLTextures();

  /* perspective */
  gl.persp.angle=45.0;
  gl.persp.aspect=1.0;
  gl.persp.zNear=0.1;
  gl.persp.zFar=100.0;
  gl.quad.sl=gl.quad.st*2;

  glClearColor(gl.bg[0],gl.bg[1],gl.bg[2],gl.bg[3]);
  glClearDepth(1.0);
  glClearStencil(0);
  glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
  glDepthFunc(GL_LESS);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glEnable(GL_DEPTH_TEST);
  glShadeModel(GL_SMOOTH);
  glViewport(0,0,gl.width,gl.height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(gl.persp.angle,gl.persp.aspect,
      gl.persp.zNear,gl.persp.zFar);
  glMatrixMode(GL_MODELVIEW);
  gl.width=gl.glarea->allocation.width;
  gl.height=gl.glarea->allocation.height;

  /* light */
  glEnable(GL_LIGHT0);
  glDisable(GL_LIGHT0);
  for (a=0;a<gl.lightmodel.model;a++) {
    glLightModelfv(gl.lightmodel.modelname[a],
		gl.lightmodel.modelparam[a]);
  }
  for (a=0;a<MaxLight;a++) {
    if (gl.light[a].on) {
      for (b=0;b<MaxLightParam;b++) {
	glLightfv(gl.light[a].light, gl.light[a].type[b],
	    gl.light[a].param[b]);
      }
      glEnable(gl.light[a].light);
    } else {
      glDisable(gl.light[a].light);
    }
  }
  glEnable(GL_LIGHTING);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_LIGHT0);

  /* Create a new quadratic for the spheres */
  gl.quad.q = gluNewQuadric();
  gluQuadricNormals(gl.quad.q, GL_SMOOTH);
  gluQuadricTexture(gl.quad.q, GL_TRUE);
  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
  Redraw_Light();
  glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
  anim.depth=0;
}

/* EOF */
